/**
 * @file generator.cpp
 * @author Jakub Kierasiński, 318377
 * @brief ZPR - zadanie dodatkowe. Przedstawienie nowej funkcjonalności dodanej w standardzie C++23: std::generator.
 *
 * Bibliografia:
 * https://en.cppreference.com/w/cpp/23
 * > https://www.open-std.org/jtc1/sc22/wg21/docs/papers/2022/p2502r2.pdf
 *
 * https://en.cppreference.com/w/cpp/coroutine/generator
 *
 * Kompilacja:
 * g++ -std=c++23 generator.cpp __generator.hpp
 */

#include <coroutine>
#include <iostream>
#include <memory>
#include <memory_resource>
#include <random>
#include <ranges>
#include <string>
//#include <generator>
#include "__generator.hpp"

/**
 * @brief Szablon klasy std::generator jest gotową do użycia korutyną (coroutine), która umożliwia generację ciągów elementów.
 * Wywołanie co_yield w ciele funkcji generuje jeden element ciągu. Ta funkcjonalność jest szczególnie użyteczna, kiedy
 * pożądane jest generowanie elementów w sposób leniwy (tzn. dopiero przy kolejnych zawołaniach funkcji).
 *
 */

/************************************************** BEGIN NUMBERS ***********************************************************/

// Definicja funkcji generatora, która zwraca sekwencję liczb Fibonacciego
std::generator<int> fibonacci()
{
    auto a = 0, b = 1;
    for (auto _ : std::views::iota(0)) {
        co_yield std::exchange(a, std::exchange(b, a + b));
    }
}

// Definicja funkcji generatora, która zwraca sekwencję kolejnych liczb nieparzystych
std::generator<int> oddNumbersGenerator()
{
    int i = 0;
    while (true) {
        if (!(i % 2)) {
            co_yield i;
        }
        ++i;
    }
}

// Definicja funkcji generatora, która zwraca sekwencję kolejnych liczb nieparzystych z ograniczeniem
std::generator<int> oddNumbersGeneratorRestricted(int r)
{
    int i = 0;
    while (i < r) {
        if (!(i % 2)) {
            co_yield i;
        }
        ++i;
    }
}

void numbersGeneratorExample()
{
    // Użycie generatora wraz z widokiem std::views::take, który ogranicza liczbę elementów do 10
    for (auto const& v : fibonacci() | std::views::take(10)) {
        std::cout << v << " "; // wypisuje elementy sekwencji na standardowe wyjście
    }
    std::cout << std::endl;

    // Generowanie kolejnych nieparzystych liczb ograniczonych przez std::views::take_while
    for (auto const& v : oddNumbersGenerator() | std::views::take_while([](int v) { return v < 10; })) {
        std::cout << v << ' ';
    }
    std::cout << std::endl;

    // Generowanie kolejnych nieparzystych liczb ograniczonych przez std::generator
    for (auto const& v : oddNumbersGeneratorRestricted(10)) {
        std::cout << v << ' ';
    }
    std::cout << std::endl;
}

/************************************************** END NUMBERS ***********************************************************/
/************************************************** BEGIN STRINGS *********************************************************/

// Funkcja generująca kolejne znaki alfabetu
std::generator<char> alphabetGenerator()
{
    char c = 'a';
    co_yield c;
    while (c < 'z') {
        ++c;
        co_yield c;
    }
}

// Możliwe jest zagnieżdżenie jednego generatora w innym
std::generator<std::string_view, std::string_view> stringViews()
{
    co_yield "foo";
    co_yield "bar";
}

std::generator<std::string_view, std::string> strings()
{
    co_yield "start";
    std::string s;
    for (auto sv : stringViews()) {
        s = sv;
        s.push_back('!');
        co_yield s;
    }
    co_yield "end";
}

void stringsExample()
{
    // Wyświetlenie kolejnych liter alfabetu angielskiego
    for (auto const& v : alphabetGenerator()) {
        std::cout << v << " ";
    }
    std::cout << std::endl;
    // Dostępne są tylko 4 obiekty i tylko one są zwracane.
    for (auto const& v : strings() | std::views::take(10)) {
        std::cout << v << " ";
    }
    std::cout << std::endl;
}

/************************************************** END STRINGS ***********************************************************/
/************************************************** BEGIN TREES ***********************************************************/
// Struktura drzewiasta
// Z pomocą std::generator można w łatwy i intuicyjny sposób rekurencyjnie iterować po całej strukturze drzewiastej
struct Tree {
    std::shared_ptr<Tree> left;
    std::shared_ptr<Tree> right;
    int value;
};

std::generator<int> visit(std::shared_ptr<Tree> tree)
{
    if (tree.get()->left)
        co_yield std::ranges::elements_of(visit(tree.get()->left));
    co_yield tree.get()->value;
    if (tree.get()->right)
        co_yield std::ranges::elements_of(visit(tree.get()->right));
}

void treesExample()
{
    Tree leftLeaf { nullptr, nullptr, 10 };
    Tree rightLeaf { nullptr, nullptr, 20 };
    Tree rightNode { nullptr, std::make_shared<Tree>(rightLeaf), 15 };
    Tree root { std::make_shared<Tree>(leftLeaf), std::make_shared<Tree>(rightNode), 1 };
    std::shared_ptr<Tree> pRoot = std::make_shared<Tree>(std::make_shared<Tree>(leftLeaf), std::make_shared<Tree>(rightNode), 1);

    for (const auto& v : visit(pRoot)) {
        std::cout << v << " ";
    }
    std::cout << std::endl;
}

/************************************************** END TREES ***********************************************************/
/************************************************** BEGIN OBJECTS *******************************************************/
// Generowanie obiektów z losowymi parametrami
class MyObject {
public:
    MyObject(int id, std::string name)
        : id_(id)
        , name_(name)
    {
    }
    int id() const { return id_; }
    std::string name() const { return name_; }

private:
    int id_;
    std::string name_;
};

// Funkcja generująca ciąg obiektów z losowymi parametrami
std::generator<MyObject> object_generator()
{
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> id_dist(1, 100);
    std::uniform_int_distribution<> name_dist(1, 10);
    while (true) {
        co_yield MyObject(id_dist(gen), std::string(name_dist(gen), 'x'));
    }
}

void objectsExample()
{
    // Wyświetlenie pierwszych 5 wygenerowanych obiektów
    for (auto const& v : object_generator() | std::views::take(5)) {
        std::cout << "ID: " << v.id() << ", Name: " << v.name() << '\n';
    }
}

/************************************************** END OBJECTS ***********************************************************/

int main()
{
    numbersGeneratorExample();
    stringsExample();
    treesExample();
    objectsExample();
}
